## Projets

**Modalités**

 - Travail en groupe, mais note individuelle
 - Rendu
   - Un rapport (pour le groupe)
   - Un journal (par personne)
   - Une présentation (par groupe)


## Groupes de projets

 - **Kubernetes/spark**
   - Capt
   - Anthonioz
   - Lavorel

 - **Kubernetes/slurm**
   - Agodomou
   - Tayan
   - Nguyen

 - **IA 1**
   - Harb
   - Marino
   - Pavlovich

 - **IA 2**
   - Bonga
   - Sanz
   - Dahmoun

 - **Performances GPU**
   - Yu
   - Burgos
   - Riser (AMD)


 - Groupe non attribué : Rimensberger

## Modules pour CUDA

Ensemble de modules permettant d'avoir CUDA, Python 3.10 et Qt (pour le profiling interactif).

```
module load CUDA/12.3.0 GCC/12.2.0 Python/3.10.8 Qt5/5.15.7
```


:construction: (à compléter)

## Profiling

Afin d'évaluer les performances d'un logiciel, le "profiling" est la méthode
plus avancée, car elle fournit un profil d'utilisation des ressources
logicielles et matérielles (SW/HW).

Dans le cadre des logiciels écrit en CUDA, deux outils sont disponibles:
  * [Nsight System](https://docs.nvidia.com/nsight-systems/UserGuide/index.html):
    profiler classique pour l'analyse globale (système) de l'application,
  * [Nsight Compute](https://docs.nvidia.com/nsight-compute/NsightComputeCli/index.html):
    profiler avancé axé sur l'optimisation de l'application.

Ces deux logiciel sont donc complémentaires et peuvent être lancés soit en
ligne de commande (CLI), soit en mode graphique (GUI). On les utilise
normalement en mode interactif, en se connectant en SSH sur un nœud de calcul
avec une GPU rattachée, p. ex. (l'option `-Y` est obligatoire!):
``` shell
lcl$ ssh -Y votre-système-hpc
hpc$ salloc --partition=shared-gpu --gpus=1 --mem=8G -t 0:15:00
gpu$ nsys ...     # lancement de la CLI, ou
gpu$ nsys-ui ...  # lancement de la GUI
```

:bulb: On peut lancer le profiling directement dans la GUI, mais il peut y
avoir de problèmes de stabilité (crash)! Donc, on recommande de profiler votre
programme en CLI et puis analyser le rapport en GUI.


### Nsight System

Aide en ligne:
``` shell
gpu$ ncu --help
gpu$ ncu profile --help
```

Profiling de l'application en CLI:
``` shell
gpu$ ncu profile YOUR_CUDA_APP [OPTIONS]
```

Exemple:
``` shell
gpu$ ncu profile vec_add_grid
...
Generating '/tmp/ncu-report-43ac.qdstrm'
[1/1] [========================100%] report1.ncu-rep
Generated:
    /home/users/.../report1.ncu-rep
```

On obtient ainsi le rapport `report1.ncu-rep` que l'on peut ensuite analyser
dans la GUI:
``` shell
gpu$ ncu-ui report1.ncu-rep
```

### Nsight Compute

Aide en ligne:
``` shell
gpu$ ncu --help
```

Profiling de l'application en CLI, avec l'option `-o` pour garder le rapport dans un fichier:
``` shell
gpu$ ncu -o report1.ncu-rep YOUR_CUDA_APP [OPTIONS]
```

Exemple:
``` shell
gpu$ ncu -o report1.ncu-rep vec_add_grid
...
==PROF== Disconnected from process 2485633
==PROF== Report: .../report1.ncu-rep
```

On obtient ainsi le rapport `report1.ncu-rep` que l'on peut ensuite analyser
dans la GUI:
``` shell
gpu$ ncu-ui report1.ncu-rep
```

<!-- Local Variables: -->
<!-- jinx-languages: "fr_FR" -->
<!-- End: -->
