#include<stdio.h>

__device__ unsigned int t_gamma(unsigned int pixel, double gamma) {
  double pixel_norm = ((double) pixel)/255.0;
  double new_pixel = 255.0*pow(pixel_norm, gamma);
  return rint(new_pixel);
}
__global__ void alpha_ker(unsigned int* img, int nRows, int nCols, double gamma) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  int k = (i*nCols) + j;
  if (i >= nRows || j >= nCols)
    return;
  img[k] = t_gamma(img[k], gamma);
}

void print_ary(unsigned int* img, int nRows, int nCols) {
  for (int i = 0; i < nRows; i++) {
    for (int j = 0; j < nCols; j++) {
      printf("%d, ", img[i*nCols + j]);
    }
    printf("\n");
  }
}

void h_gamma(unsigned int* img, int nRows, int nCols, double gamma){
  for(int i=0; i<nRows; i++){
    for(int j=0; j<nCols; j++){
        double pixel_norm = ((double) img[i*nCols+j])/255.0;
        double new_pixel = 255.0*pow(pixel_norm, gamma);
        img[i*nCols+j] = rint(new_pixel);
    }
  }
}

int main() {
  double gamma = 10.0;
  int NUM_ROWS = 20;
  int NUM_COLS = 20;
  int imgSize = NUM_ROWS*NUM_COLS*sizeof(unsigned int);

  unsigned int* h_img = (unsigned int*) malloc(imgSize);
  unsigned int* h_img_res = (unsigned int*) malloc(imgSize);
  unsigned int* d_img;
  
  //cudaError_t cuError = cudaSuccess;  
  srand(10);
  for (int i = 0; i < NUM_ROWS; i++) {
    for (int j = 0; j < NUM_COLS; j++) {
      h_img[i*NUM_COLS + j] = ((double)rand()/RAND_MAX)*255;
    }
  }
  memcpy(h_img_res, h_img, imgSize);

  print_ary(h_img_res, NUM_ROWS, NUM_COLS);

  cudaMalloc(&d_img, imgSize);
  cudaMemcpy(d_img, h_img_res, imgSize, cudaMemcpyHostToDevice);
  
  //cuError = cudaMalloc((void**)&da, iSize);
  //if (cudaSuccess != cuError) {
  //  printf("Failed to allocate memory\n");
  //  return 1;
  //}
  //cuError = cudaMemcpy(da, ha, iSize, cudaMemcpyHostToDevice);

  dim3 dimGrid (2, 2, 1);
  dim3 dimBlock (10, 10, 1);

  
  alpha_ker<<<dimGrid, dimBlock>>>(d_img, NUM_ROWS, NUM_COLS, gamma);

  //cuError = cudaGetLastError();
  cudaMemcpy(h_img_res, d_img, imgSize, cudaMemcpyDeviceToHost);
  cudaFree(d_img) ;

  printf("======================================\n");

  print_ary(h_img_res, NUM_ROWS, NUM_COLS);

  h_gamma(h_img, NUM_ROWS, NUM_COLS, gamma);

  if(memcmp(h_img, h_img_res, imgSize)==0){
    printf("tout va bien\n");
    return 0;
  }
  printf("argl!\n");
  return 1;
}